
#ifndef __nanojit_Instrinsics__
#define __nanojit_Instrinsics__

#if defined(_MSC_VER)
#define ROTL32(x,y)     _rotl(x, y)
#define ROTL64(x,y)     _rotl64(x, y)
#define BIG_CONSTANT(x) (x)
#else
#define ROTL32(x,y)     ((x << y) | (x >> (32 - y)))
#define ROTL64(x,y)     ((x << y) | (x >> (64 - y)))
#define BIG_CONSTANT(x) (x##LLU)
#endif

#if defined(_MSC_VER)
#define BYTESWAP32(x) _byteswap_ulong((x))
#define BYTESWAP64(x) _byteswap_uint64((x))
#define POPCOUNT32(x) __popcnt((x))
#define POPCOUNT64(x) __popcnt64((x))
#elif defined(__ghs__)

namespace nanojit
{

	inline uint32_t byteSwap32(uint32_t x)
	{
		return ( (((x) & 0xFF) << 24) | ((((x) >> 8) & 0xFF) << 16) | ((((x) >> 16) & 0xFF) << 8) | ((x) >> 24) );
	}
	inline uint64_t byteSwap64(uint64_t x)
	{
		return ( ((((x) >> 56) & 0xFF) <<  0) | ((((x) >> 48) & 0xFF) <<  8) | ((((x) >> 40) & 0xFF) << 16) | ((((x) >> 32) & 0xFF) << 24) | ((((x) >> 24) & 0xFF) << 32) | ((((x) >> 16) & 0xFF) << 40) | ((((x) >>  8) & 0xFF) << 48) | ((((x) >>  0) & 0xFF) << 56) );
	}
	inline unsigned int popCount32(uint32_t v)
	{
		// http://graphics.stanford.edu/~seander/bithacks.html#CountBitsSetParallel
		v = v - ((v >> 1) & 0x55555555);                    
		v = (v & 0x33333333) + ((v >> 2) & 0x33333333);     
		return ((v + (v >> 4) & 0xF0F0F0F) * 0x1010101) >> 24;
	}
	inline unsigned int popCount64(uint64_t x)
	{
		// http://graphics.stanford.edu/~seander/bithacks.html#CountBitsSetParallel
		x -= (x >> 1) & 0x5555555555555555ULL;
		x = (x & 0x3333333333333333ULL) + ((x >> 2) & 0x3333333333333333ULL);
		x = (x + (x >> 4)) & 0x0f0f0f0f0f0f0f0fULL;
		return int((x * 0x0101010101010101ULL) >> 56);
	}

}

#define BYTESWAP32(x) ::nanojit::byteSwap32(x)
#define BYTESWAP64(x) ::nanojit::byteSwap64(x)
#define POPCOUNT32(x) ::nanojit::popCount32(x)
#define POPCOUNT64(x) ::nanojit::popCount64(x)

#elif defined(__GNUC__) && __GNUC_PREREQ(4,3)
#define BYTESWAP32(x) __builtin_bswap32((x))
#define BYTESWAP64(x) __builtin_bswap64((x))
#define POPCOUNT32(x) __builtin_popcountl((x))
#define POPCOUNT64(x) __builtin_popcountll((x))
#else

namespace nanojit
{

	inline uint32_t byteSwap32(uint32_t x)
	{
		return ( (((x) & 0xFF) << 24) | ((((x) >> 8) & 0xFF) << 16) | ((((x) >> 16) & 0xFF) << 8) | ((x) >> 24) );
	}
	inline uint64_t byteSwap64(uint64_t x)
	{
		return ( ((((x) >> 56) & 0xFF) <<  0) | ((((x) >> 48) & 0xFF) <<  8) | ((((x) >> 40) & 0xFF) << 16) | ((((x) >> 32) & 0xFF) << 24) | ((((x) >> 24) & 0xFF) << 32) | ((((x) >> 16) & 0xFF) << 40) | ((((x) >>  8) & 0xFF) << 48) | ((((x) >>  0) & 0xFF) << 56) );
	}
	inline unsigned int popCount32(uint32_t v)
	{
		// http://graphics.stanford.edu/~seander/bithacks.html#CountBitsSetParallel
		v = v - ((v >> 1) & 0x55555555);                    
		v = (v & 0x33333333) + ((v >> 2) & 0x33333333);     
		return ((v + (v >> 4) & 0xF0F0F0F) * 0x1010101) >> 24;
	}
	inline unsigned int popCount64(uint64_t x)
	{
		// http://graphics.stanford.edu/~seander/bithacks.html#CountBitsSetParallel
		x -= (x >> 1) & 0x5555555555555555ULL;
		x = (x & 0x3333333333333333ULL) + ((x >> 2) & 0x3333333333333333ULL);
		x = (x + (x >> 4)) & 0x0f0f0f0f0f0f0f0fULL;
		return int((x * 0x0101010101010101ULL) >> 56);
	}

}

#define BYTESWAP32(x) ::nanojit::byteSwap32(x)
#define BYTESWAP64(x) ::nanojit::byteSwap64(x)
#define POPCOUNT32(x) ::nanojit::popCount32(x)
#define POPCOUNT64(x) ::nanojit::popCount64(x)

#endif

// countLeadingZeroes / countTrailingZeroes

#if defined(_WIN32) && (_MSC_VER >= 1300) && (defined(_M_IX86) || defined(_M_AMD64) || defined(_M_X64))

extern "C" unsigned char _BitScanForward(unsigned long * Index, unsigned long Mask);
extern "C" unsigned char _BitScanReverse(unsigned long * Index, unsigned long Mask);
#pragma intrinsic(_BitScanForward)
#pragma intrinsic(_BitScanReverse)

namespace nanojit
{

	inline unsigned countLeadingZeroes32(uint32_t x) {
		unsigned long idx;
		_BitScanReverse(&idx, (unsigned long)x);
		return idx;
	}

	inline unsigned countTrailingZeroes32(uint32_t x) {
		unsigned long idx;
		_BitScanForward(&idx, (unsigned long)x);
		return idx;
	}

}

// use 64 bits intrinsics if available
#if defined(_M_AMD64) || defined(_M_X64)
extern "C" unsigned char _BitScanForward64(unsigned long * Index, unsigned __int64 Mask);
extern "C" unsigned char _BitScanReverse64(unsigned long * Index, unsigned __int64 Mask);
#pragma intrinsic(_BitScanForward64)
#pragma intrinsic(_BitScanReverse64)

namespace nanojit
{

	inline unsigned countLeadingZeroes64(uint64_t x)
	{
		unsigned long idx;
		_BitScanReverse64(&idx, (unsigned __int64)x);
		return idx;
	}

	inline unsigned countTrailingZeroes64(uint64_t x)
	{
		unsigned long idx;
		_BitScanForward64(&idx, (unsigned __int64)x);
		return idx;
	}

}

#else

inline unsigned countLeadingZeroes64(uint64_t x) {
	return (x & 0xffffffff00000000LL) ? countLeadingZeroes32(uint32_t(x >> 32)) + 32 : countLeadingZeroes32(uint32_t(x));
}

inline unsigned countTrailingZeroes64(uint64_t x) {
	return (x & 0x00000000ffffffffLL) ? countTrailingZeroes32(uint32_t(x)) : countTrailingZeroes32(uint32_t(x >> 32)) + 32;
}

#endif

#elif defined(__ghs__)

#include <ppc_ghs.h>

namespace nanojit
{

	__attribute__((always_inline)) unsigned countLeadingZeroes32(uint32_t x)
	{
		return __CLZ32(x);
	}

	__attribute__((always_inline)) unsigned countTrailingZeroes32(uint32_t x)
	{
		uint32_t x0 = x;
		uint32_t x1 = x - 1;
		uint32_t x2 = x1 & ~x0; // should be compiled as andc
		uint32_t x3 = __CLZ32(x2);
		uint32_t x4 = 32 - x3;
		return x4;
	}

	__attribute__((always_inline)) unsigned countLeadingZeroes64(uint64_t x)
	{
		return __CLZ64(x);
	}

	__attribute__((always_inline)) unsigned countTrailingZeroes64(uint64_t x)
	{
		uint64_t x0 = x;
		uint64_t x1 = x - 1;
		uint64_t x2 = x1 & ~x0; // should be compiled as andc
		uint64_t x3 = __CLZ64(x2);
		uint64_t x4 = 64 - x3;
		return x4;
	}
}

#elif (__GNUC__ >= 4) || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4)

namespace nanojit
{

	inline unsigned countLeadingZeroes32(uint32_t x)
	{
		return __builtin_clz(x);
	}

	inline unsigned countTrailingZeroes32(uint32_t x) {
		return __builtin_ctz(x);
	}

	inline unsigned countLeadingZeroes64(uint64_t x)
	{
		return __builtin_clzll(x);
	}

	// Returns the index of the least significant bit that is set.
	inline unsigned countTrailingZeroes64(uint64_t x)
	{
		return __builtin_ctzll(x);
	}

}

#else

namespace nanojit
{

	template<typename T, unsigned int BIT_COUNT>
	inline unsigned int softCountLeadingZeroes(T x)
	{
		T iMask = static_cast<T>(1) << (BIT_COUNT - 1);
		unsigned int iCountLeadingZeroes = 0;
		while ((iCountLeadingZeroes < BIT_COUNT) && ((x & iMask) == 0))
		{
			iCountLeadingZeroes++;
			iMask >>= 1;
		}
		return iCountLeadingZeroes;
	}

	template<typename T, unsigned int BIT_COUNT>
	inline unsigned int softCountTrailingZeroes(T x)
	{
		T iMask = 1;
		unsigned int iCountTrailingZeroes = 0;
		while ((iCountTrailingZeroes < BIT_COUNT) && ((x & iMask) == 0))
		{
			iCountTrailingZeroes++;
			iMask <<= 1;
		}
		return iCountTrailingZeroes;
	}

	inline unsigned countLeadingZeroes32(uint32_t x)
	{
		return softCountLeadingZeroes<uint32_t, 32>(x);
	}

	inline unsigned countTrailingZeroes32(uint32_t x)
	{
		return softCountTrailingZeroes<uint32_t, 32>(x);
	}

	inline unsigned countLeadingZeroes64(uint64_t x)
	{
		return softCountLeadingZeroes<uint64_t, 64>(x);
	}

	inline unsigned countTrailingZeroes64(uint64_t x)
	{
		return softCountTrailingZeroes<uint64_t, 64>(x);
	}
}

#endif 

namespace nanojit
{

	/**
	 * @brief Clears the less significant bit of a integer
	 */
	template< typename T >
	NANOJIT_FORCEINLINE T clearLeastSignificantBit(T a_iValue)
	{
		T iValue = a_iValue;
		iValue &= ~( iValue & ~(iValue - 1) );
		return iValue;
	}

	NANOJIT_FORCEINLINE uint32_t countLeadingZeroes(uint32_t x)
	{
		return countLeadingZeroes32(x);
	}

	NANOJIT_FORCEINLINE uint64_t countLeadingZeroes(uint64_t x)
	{
		return countLeadingZeroes64(x);
	}

	NANOJIT_FORCEINLINE uint32_t countTrailingZeroes(uint32_t x)
	{
		return countTrailingZeroes32(x);
	}

	NANOJIT_FORCEINLINE uint64_t countTrailingZeroes(uint64_t x)
	{
		return countTrailingZeroes64(x);
	}

	NANOJIT_FORCEINLINE uint32_t clearLeastSignificantBit(uint32_t x)
	{
		return clearLeastSignificantBit<uint32_t>(x);
	}

	NANOJIT_FORCEINLINE uint64_t clearLeastSignificantBit(uint64_t x)
	{
		return clearLeastSignificantBit<uint64_t>(x);
	}
}


#endif // __nanojit_Instrinsics__

